@@ -1,20 +1,7 @@
 /* This file is part of the KDE project
  * Copyright 2008 (C) Boudewijn Rempt <boud@valdyas.org>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public License
- * along with this library; see the file COPYING.LIB.  If not, write to
- * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
- * Boston, MA 02110-1301, USA.
+ * SPDX-License-Identifier: LGPL-2.0-or-later
  */
 #include "kis_kra_saver.h"
 
@@ -39,11 +26,10 @@
 #include <KoColorProfile.h>
 #include <KoColor.h>
 #include <KoColorSet.h>
-#include <KoResourceServer.h>
-#include <KoResourceServerProvider.h>
 #include <KoStore.h>
 #include <KoStoreDevice.h>
-
+#include <KisResourceTypes.h>
+#include <KisResourceModel.h>
 #include <kis_annotation.h>
 #include <kis_image.h>
 #include <kis_image_animation_interface.h>
@@ -53,16 +39,16 @@
 #include <kis_adjustment_layer.h>
 #include <kis_layer_composition.h>
 #include <kis_painting_assistants_decoration.h>
-#include <kis_psd_layer_style_resource.h>
 #include "kis_png_converter.h"
 #include "kis_keyframe_channel.h"
-#include <kis_time_range.h>
+#include <kis_time_span.h>
 #include "KisDocument.h"
 #include <string>
 #include "kis_dom_utils.h"
 #include "kis_grid_config.h"
 #include "kis_guides_config.h"
 #include "KisProofingConfiguration.h"
+#include "kis_asl_layer_style_serializer.h"
 
 #include <KisMirrorAxisConfig.h>
 
@@ -74,20 +60,30 @@ using namespace KRA;
 
 struct KisKraSaver::Private
 {
-public:
     KisDocument* doc;
     QMap<const KisNode*, QString> nodeFileNames;
     QMap<const KisNode*, QString> keyframeFilenames;
     QString imageName;
     QString filename;
     QStringList errorMessages;
+    QStringList warningMessages;
+    QStringList specialAnnotations;
+    bool addMergedImage;
+    QList<KoResourceLoadResult> linkedDocumentResources;
+
+    Private() {
+        specialAnnotations << "exif" << "icc";
+    }
+
 };
 
-KisKraSaver::KisKraSaver(KisDocument* document, const QString &filename)
-        : m_d(new Private)
+KisKraSaver::KisKraSaver(KisDocument* document, const QString &filename, bool addMergedImage)
+    : m_d(new Private)
 {
     m_d->doc = document;
     m_d->filename = filename;
+    m_d->addMergedImage = addMergedImage;
+    m_d->linkedDocumentResources = document->linkedDocumentResources();
 
     m_d->imageName = m_d->doc->documentInfo()->aboutInfo("title");
     if (m_d->imageName.isEmpty()) {
@@ -129,7 +125,7 @@ QDomElement KisKraSaver::saveXML(QDomDocument& doc,  KisImageSP image)
     }
 
     quint32 count = 1; // We don't save the root layer, but it does count
-    KisSaveXmlVisitor visitor(doc, imageElement, count, m_d->doc->url().toLocalFile(), true);
+    KisSaveXmlVisitor visitor(doc, imageElement, count, m_d->filename, true);
     visitor.setSelectedNodes({m_d->doc->preActivatedNode()});
 
     image->rootLayer()->accept(visitor);
@@ -147,62 +143,234 @@ QDomElement KisKraSaver::saveXML(QDomDocument& doc,  KisImageSP image)
     saveGuides(doc, imageElement);
     saveMirrorAxis(doc, imageElement);
     saveAudio(doc, imageElement);
-    savePalettesToXML(doc, imageElement);
+    saveResourcesToXML(doc, imageElement);
 
+    // Redundancy -- Save animation metadata in XML to prevent data loss for the time being...
     QDomElement animationElement = doc.createElement("animation");
     KisDomUtils::saveValue(&animationElement, "framerate", image->animationInterface()->framerate());
     KisDomUtils::saveValue(&animationElement, "range", image->animationInterface()->fullClipRange());
     KisDomUtils::saveValue(&animationElement, "currentTime", image->animationInterface()->currentUITime());
     imageElement.appendChild(animationElement);
 
+    vKisAnnotationSP_it beginIt = image->beginAnnotations();
+    vKisAnnotationSP_it endIt = image->endAnnotations();
+
+    if (beginIt != endIt) {
+        QDomElement annotationsElement = doc.createElement(ANNOTATIONS);
+        vKisAnnotationSP_it it = beginIt;
+        while (it != endIt) {
+            if (!(*it) || (*it)->type().isEmpty()) {
+                it++;
+                continue;
+            }
+            QString type = (*it)->type();
+
+            if (!m_d->specialAnnotations.contains(type)) {
+
+                QString description = (*it)->description();
+                QDomElement annotationElement = doc.createElement(ANNOTATION);
+                annotationsElement.appendChild(annotationElement);
+                annotationElement.setAttribute("type", type);
+                annotationElement.setAttribute("description", description);
+            }
+            it++;
+        }
+        imageElement.appendChild(annotationsElement);
+    }
+
+
     return imageElement;
 }
 
-bool KisKraSaver::savePalettes(KoStore *store, KisImageSP image, const QString &uri)
+bool KisKraSaver::saveResources(KoStore *store, KisImageSP image, const QString &uri)
 {
     Q_UNUSED(image);
     Q_UNUSED(uri);
 
-    bool res = false;
-    if (m_d->doc->paletteList().size() == 0) {
-        return true;
-    }
-    for (const KoColorSet *palette : m_d->doc->paletteList()) {
-        if (!palette->isGlobal()) {
-            if (!store->open(m_d->imageName + PALETTE_PATH + palette->filename())) {
-                m_d->errorMessages << i18nc("Error message when saving a .kra file", "Could not save palettes.");
-                return false;
-            }
-            QByteArray ba = palette->toByteArray();
-            qint64 nwritten = 0;
-            if (!ba.isEmpty()) {
-                nwritten = store->write(ba);
-            } else {
-                qWarning() << "Cannot save the palette to a byte array:" << palette->name();
-            }
-            res = store->close();
-            res = res && (nwritten == ba.size());
+    QList<KoResourceLoadResult> embeddedResources = m_d->linkedDocumentResources;
+
+    Q_FOREACH (const KoResourceLoadResult &result, embeddedResources) {
+        KIS_SAFE_ASSERT_RECOVER(result.type() != KoResourceLoadResult::ExistingResource) { continue; }
+
+        if (result.type() == KoResourceLoadResult::FailedLink) {
+            m_d->warningMessages << i18nc("Error message when saving a .kra file", "Could not export resource for embedding: %1", result.signature().filename);
+            continue;
+        }
+
+        KoEmbeddedResource resource = result.embeddedResource();
+
+        QString path = RESOURCE_PATH + "/" + resource.signature().type;
+
+        if (resource.signature().type == ResourceType::Palettes) {
+            path = m_d->imageName + PALETTE_PATH;
+        }
+
+        const QString fileName = resource.signature().filename;
+
+        if (!store->open(path  + '/' + fileName)) {
+            m_d->warningMessages << i18nc("Error message when saving a .kra file", "Could not write resource: %1", result.signature().filename);
+            continue;
+        }
+
+        // we first read into a buffer to make sure the save operation is transactional,
+        // that is, either resource is saves correctly, or the file is left empty.
+        QByteArray ba = resource.data();
+
+        qint64 nwritten = 0;
+        if (!ba.isEmpty()) {
+            nwritten = store->write(ba);
+        } else {
+            m_d->warningMessages << i18nc("Error message when saving a .kra file", "Written resource is empty: %1", result.signature().filename);
+        }
+
+        store->close();
+
+        if (nwritten != ba.size()) {
+            m_d->warningMessages << i18nc("Error message when saving a .kra file", "Written resource is incomplete: %1", result.signature().filename);
         }
     }
 
-    if (!res) {
-        m_d->errorMessages << i18nc("Error message when saving a .kra file", "Could not save palettes.");
-    }
-    return res;
+    return true;
 }
 
+bool KisKraSaver::saveStoryboard(KoStore *store, KisImageSP image, const QString &uri)
+{
+    Q_UNUSED(image);
+    Q_UNUSED(uri);
 
-void KisKraSaver::savePalettesToXML(QDomDocument &doc, QDomElement &element)
+    bool success = true;
+    if (m_d->doc->getStoryboardItemList().count() == 0) {
+        return true;
+    } else {
+        if (!store->open(m_d->imageName + STORYBOARD_PATH + "index.xml")) {
+            m_d->errorMessages << i18nc("Error message when saving a .kra file", "Could not save storyboards.");
+            return false;
+        }
+
+        QDomDocument storyboardDocument = m_d->doc->createDomDocument("storyboard-info", "1.1");
+        QDomElement root = storyboardDocument.documentElement();
+        saveStoryboardToXML(storyboardDocument, root);
+
+        QByteArray ba = storyboardDocument.toByteArray();
+        qint64 nwritten = 0;
+        if (!ba.isEmpty()) {
+            nwritten = store->write(ba);
+        } else {
+            success = false;
+            qWarning() << "Could not save storyboard data to a byte array!";
+        }
+
+        bool r = store->close();
+        success = success && r && (nwritten == ba.size());
+    }
+
+    if (!success) {
+        m_d->errorMessages << i18nc("Error message when saving a .kra file", "Could not save storyboards.");
+        return false;
+    }
+
+    return success;
+}
+
+bool KisKraSaver::saveAnimationMetadata(KoStore *store, KisImageSP image, const QString &uri)
+{
+    Q_UNUSED(uri);
+
+    if (!store->open(m_d->imageName + ANIMATION_METADATA_PATH + "index.xml")) {
+        m_d->errorMessages << i18nc("Error message when saving a .kra file", "Could not save animation meta data.");
+        return false;
+    }
+
+    QDomDocument animationDocument = m_d->doc->createDomDocument("animation-metadata", "1.1");
+    QDomElement root = animationDocument.documentElement();
+    saveAnimationMetadataToXML(animationDocument, root, image);
+
+    bool success = true;
+
+    QByteArray ba = animationDocument.toByteArray();
+    qint64 nwritten = 0;
+    if (!ba.isEmpty()) {
+        nwritten = store->write(ba);
+    } else {
+        qWarning() << "Could not save animation meta data to a byte array!";
+        success = false;
+    }
+
+    bool r = store->close();
+
+    success = success && r && (nwritten == ba.size());
+
+    if (!success) {
+        m_d->errorMessages << i18nc("Error message when saving a .kra file", "Could not save animation meta data.");
+        return false;
+    }
+
+    return true;
+}
+
+void KisKraSaver::saveResourcesToXML(QDomDocument &doc, QDomElement &element)
 {
     QDomElement ePalette = doc.createElement(PALETTES);
-    for (const KoColorSet *palette : m_d->doc->paletteList()) {
-        if (!palette->isGlobal()) {
-            QDomElement eFile =  doc.createElement("palette");
-            eFile.setAttribute("filename", palette->filename());
-            ePalette.appendChild(eFile);
+    QDomElement eResources = doc.createElement(RESOURCES);
+
+    Q_FOREACH (const KoResourceLoadResult resource, m_d->linkedDocumentResources) {
+        // all warnings will be issued in KisKraSaver::saveResources()
+        if (resource.type() != KoResourceLoadResult::EmbeddedResource) continue;
+
+        KoResourceSignature sig = resource.signature();
+
+        QDomElement eResource = doc.createElement("resource");
+        eResource.setAttribute("type", sig.type);
+        eResource.setAttribute("name", sig.name);
+        eResource.setAttribute("filename", sig.filename);
+        eResource.setAttribute("md5sum", sig.md5sum);
+
+        if (sig.type == ResourceType::Palettes) {
+            ePalette.appendChild(eResource);
+        }
+        else {
+            eResources.appendChild(eResource);
+
         }
     }
     element.appendChild(ePalette);
+    element.appendChild(eResources);
+}
+
+void KisKraSaver::saveStoryboardToXML(QDomDocument& doc, QDomElement &element)
+{
+    //saving storyboard comments
+    QDomElement eCommentList = doc.createElement("StoryboardCommentList");
+    for (StoryboardComment comment: m_d->doc->getStoryboardCommentsList()) {
+        QDomElement commentElement = doc.createElement("storyboardcomment");
+        commentElement.setAttribute("name", comment.name);
+        commentElement.setAttribute("visibility", comment.visibility);
+        eCommentList.appendChild(commentElement);
+    }
+    element.appendChild(eCommentList);
+
+    //saving storyboard items
+    QDomElement eItemList = doc.createElement("StoryboardItemList");
+    for (StoryboardItemSP item : m_d->doc->getStoryboardItemList()) {
+        QDomElement eItem =  item->toXML(doc);
+        eItemList.appendChild(eItem);
+    }
+    element.appendChild(eItemList);
+}
+
+void KisKraSaver::saveAnimationMetadataToXML(QDomDocument &doc, QDomElement &element, KisImageSP image)
+{
+    KisDomUtils::saveValue(&element, "framerate", image->animationInterface()->framerate());
+    KisDomUtils::saveValue(&element, "range", image->animationInterface()->fullClipRange());
+    KisDomUtils::saveValue(&element, "currentTime", image->animationInterface()->currentUITime());
+
+    {
+        QDomElement exportItemElem = doc.createElement("export-settings");
+        KisDomUtils::saveValue(&exportItemElem, "sequenceFilePath", image->animationInterface()->exportSequenceFilePath());
+        KisDomUtils::saveValue(&exportItemElem, "sequenceBaseName", image->animationInterface()->exportSequenceBaseName());
+        KisDomUtils::saveValue(&exportItemElem, "sequenceInitialFrameNumber", image->animationInterface()->exportInitialFrameNumber());
+        element.appendChild(exportItemElem);
+    }
 }
 
 bool KisKraSaver::saveKeyframes(KoStore *store, const QString &uri, bool external)
@@ -253,7 +421,7 @@ bool KisKraSaver::saveNodeKeyframes(KoStore *store, QString location, const KisN
     return true;
 }
 
-bool KisKraSaver::saveBinaryData(KoStore* store, KisImageSP image, const QString &uri, bool external, bool autosave)
+bool KisKraSaver::saveBinaryData(KoStore* store, KisImageSP image, const QString &uri, bool external, bool addMergedImage)
 {
     QString location;
 
@@ -275,9 +443,6 @@ bool KisKraSaver::saveBinaryData(KoStore* store, KisImageSP image, const QString
     qint64 nwritten = 0;
 
     // saving annotations
-    // XXX this only saves EXIF and ICC info. This would probably need
-    // a redesign of the dtd of the krita file to do this more generally correct
-    // e.g. have <ANNOTATION> tags or so.
     bool savingAnnotationsSuccess = true;
     KisAnnotationSP annotation = image->annotation("exif");
     if (annotation) {
@@ -362,19 +527,54 @@ bool KisKraSaver::saveBinaryData(KoStore* store, KisImageSP image, const QString
 
     success = success && savingSoftproofingProfileSuccess;
 
+    // Save the remaining annotations
+    vKisAnnotationSP_it beginIt = image->beginAnnotations();
+    vKisAnnotationSP_it endIt = image->endAnnotations();
+
+    bool savingRemainingAnnotationsSuccess = true;
+    if (beginIt != endIt) {
+        vKisAnnotationSP_it it = beginIt;
+        while (it != endIt) {
+            if (!(*it) || (*it)->type().isEmpty()) {
+                it++;
+                continue;
+            }
+            QString type = (*it)->type();
+
+            if (!m_d->specialAnnotations.contains(type)) {
+                location = external ? QString() : uri;
+                location += m_d->imageName + ANNOTATIONS_PATH + type;
+                if (store->open(location)) {
+                    nwritten = store->write((*it)->annotation());
+                    r = store->close();
+                    savingRemainingAnnotationsSuccess = savingRemainingAnnotationsSuccess && (nwritten == (*it)->annotation().size()) && r;
+                } else {
+                    savingRemainingAnnotationsSuccess = false;
+                }
+            }
+            it++;
+        }
+    }
+
+    if (!savingRemainingAnnotationsSuccess) {
+        m_d->errorMessages.append(i18nc("Saving .kra file error message", "Could not save additional annotations."));
+    }
+
+    success = success && savingRemainingAnnotationsSuccess;
+
     bool savingLayerStylesSuccess = true;
     {
-        KisPSDLayerStyleCollectionResource collection("not-nexists.asl");
-        KIS_ASSERT_RECOVER_NOOP(!collection.valid());
-        collection.collectAllLayerStyles(image->root());
-        if (collection.valid()) {
+        KisAslLayerStyleSerializer serializer;
+        QVector<KisPSDLayerStyleSP> stylesClones = serializer.collectAllLayerStyles(image->root());
+        if (stylesClones.size() > 0) {
             location = external ? QString() : uri;
             location += m_d->imageName + LAYER_STYLES_PATH;
 
             if (store->open(location)) {
                 QBuffer aslBuffer;
                 if (aslBuffer.open(QIODevice::WriteOnly)) {
-                    collection.saveToDevice(&aslBuffer);
+                    serializer.setStyles(stylesClones);
+                    serializer.saveToDevice(&aslBuffer);
                     aslBuffer.close();
                     nwritten = store->write(aslBuffer.buffer());
                     savingLayerStylesSuccess = savingLayerStylesSuccess && (nwritten == aslBuffer.buffer().size());
@@ -396,9 +596,7 @@ bool KisKraSaver::saveBinaryData(KoStore* store, KisImageSP image, const QString
     success = success && savingLayerStylesSuccess;
 
     bool savingMergedImageSuccess = true;
-
-
-    if (!autosave) {
+    if (addMergedImage) {
         KisPaintDeviceSP dev = image->projection();
         store->setCompressionEnabled(false);
         r = KisPNGConverter::saveDeviceToStore("mergedimage.png", image->bounds(), image->xRes(), image->yRes(), dev, store);
@@ -418,11 +616,18 @@ bool KisKraSaver::saveBinaryData(KoStore* store, KisImageSP image, const QString
     return success;
 }
 
+
+
 QStringList KisKraSaver::errorMessages() const
 {
     return m_d->errorMessages;
 }
 
+QStringList KisKraSaver::warningMessages() const
+{
+    return m_d->warningMessages;
+}
+
 void KisKraSaver::saveBackgroundColor(QDomDocument& doc, QDomElement& element, KisImageSP image)
 {
     QDomElement e = doc.createElement(CANVASPROJECTIONCOLOR);
@@ -502,7 +707,7 @@ bool KisKraSaver::saveAssistants(KoStore* store, QString uri, bool external)
 
 bool KisKraSaver::saveAssistantsList(QDomDocument& doc, QDomElement& element)
 {
-    int count_ellipse = 0, count_perspective = 0, count_ruler = 0, count_vanishingpoint = 0,count_infiniteruler = 0, count_parallelruler = 0, count_concentricellipse = 0, count_fisheyepoint = 0, count_spline = 0;
+    int count_ellipse = 0, count_twopoint = 0, count_perspective = 0, count_ruler = 0, count_vanishingpoint = 0,count_infiniteruler = 0, count_parallelruler = 0, count_concentricellipse = 0, count_fisheyepoint = 0, count_spline = 0;
     QList<KisPaintingAssistantSP> assistants =  m_d->doc->assistants();
     if (!assistants.isEmpty()) {
         QDomElement assistantsElement = doc.createElement("assistants");
@@ -539,6 +744,10 @@ bool KisKraSaver::saveAssistantsList(QDomDocument& doc, QDomElement& element)
                 assist->saveXmlList(doc, assistantsElement, count_fisheyepoint);
                 count_fisheyepoint++;
             }
+            else if (assist->id() == "two point"){
+                assist->saveXmlList(doc, assistantsElement, count_twopoint);
+                count_twopoint++;
+            }
             else if (assist->id() == "ruler"){
                 assist->saveXmlList(doc, assistantsElement, count_ruler);
                 count_ruler++;
